# yapf

from pprint import pformat
import json
import jsonpickle  # type: ignore
import os
import funcy

from typing import Any


def expose(x: Any, shrink_right: int = 1) -> str:
    """Abuses modules `json` and `jsonpickle` to expose internals of
    (almost) any object.

    """
    return json.loads(jsonpickle.dumps(x))


def look(x: Any, shrink_right: int = 1) -> None:
    """Dumps any object to stdout using the technique in `expose()`.

    The parameter `shrink_right` is used to set narrowing of
    indentation. (Use `0` to turn off).

    """
    print(pformat(expose(x, shrink_right)).replace('    ', shrink_right * ' '))


def guessdesc(value: Any) -> str:
    "Guesstimate a plausible description of a value."
    denoms = funcy.compact([
        getattr(value, 'name', ''),
        getattr(value, 'desc', ''),
        getattr(value, 'id', '')
    ])
    named = denoms[0] if len(denoms) == 1 else ''
    spec = value.__class__.__name__
    if named:
        desc = f'{named}:{spec}'
        return desc
    else:
        desc = f'{value!r}:{spec}'
        return desc


def genabneuyaml(cursor: Any, ident=0, name=None, maxrecur=100,
                 spacing='  ') -> None:
    "Generate lines of 'AbneuYAML' (see `abneuyaml()` below)."
    spaces = spacing * ident
    before = spaces
    desc = guessdesc(cursor)
    if name:
        desc = f"{name}={desc}"
    yield before, desc
    if ident > maxrecur:
        print(f"abneuyaml: high recursion ({maxrecur}) reached, "
              "circular structure? at {desc}")
    if hasattr(cursor, '__dict__'):
        ident = ident + 1
        for attr, value in cursor.__dict__.items():
            for before, desc in genabneuyaml(value,
                                             ident=ident,
                                             name=attr,
                                             maxrecur=maxrecur):
                yield before, desc


def abneuyaml(cursor: Any, ident=0, name=None, maxrecur=100) -> None:
    """Encode any object in a format that is 'almost, but not entirely
    unlike YAML'

    """
    return os.linesep.join(
        [''.join(enc) for enc in genabneuyaml(cursor, maxrecur=maxrecur)])



# yapf

import hashlib
import fnmatch

from . import lang
from dataclasses import dataclass
from functools import singledispatch

import funcy


@dataclass
class Attr:
    infer_types = {}
    PrimType = lang.Undefined
    __bind__ = []

    @classmethod
    def create(cls, name, value):

        obj = cls.__new__(cls, value)
        obj.name = name

        @singledispatch
        def eq(other):
            raise NotImplementedError("Attr: eq not implemented for {}".format(
                type(other)))

        @eq.register(Attr)
        def _(other):
            "Does _"
            return other.name == obj.name and other == obj

        @eq.register(cls.PrimType)
        def _(other):
            "Does _"
            return other == obj.raw_value

        obj.eq = eq

        lang.bind_methods(cls, obj)

        return obj

    def __init__(self, name, value):
        self.name = name

    @staticmethod
    def isa(other):
        return isinstance(other, Attr.__class__)

    def sibling(self, other):
        return isinstance(other.__class__, Attr.__class__)

    def eq(self, other):
        return self.eq(self, other)

    @classmethod
    def infer(cls, name, value):
        PrimType = Attr.infer_types[type(value)]
        return PrimType.create(name, value)

    @property
    def raw_value(self):
        "Cast value to primitive type."

        # NB: Special case to avoid infinite recursion.
        if self.PrimType is str:
            return str.__str__(self)

        elif self.PrimType is lang.Undefined:
            return '????'

        else:
            return self.PrimType(self)

    @property
    def type_name(self):
        "Return name of primitive type"

        # NB: Special case to avoid infinite recursion.
        if self.PrimType is str:
            return 'str'

        else:
            return self.PrimType.__name__

    def __str__(self):
        return "{}={}:{}".format(self.name, self.raw_value, self.type_name)

    def __repr__(self):
        return str(self)

    def __hash__(self):
        "Does __hash__"
        value = self.PrimType(self)
        return hashlib.sha1('{}:{}'.format(self.name, value)).hexdigest()

    def __eq__(self, other):
        if hasattr(other, 'name'):
            return (other.name == self.name
                    and other.raw_value == self.raw_value)
        else:
            return other == self.raw_value


for name, PrimType in (('IntAttr', int),
                       ('FloatAttr', float),
                       ('TupleAttr', tuple),
                       ('ListAttr', list),
                       ('StrAttr', str)):  # yapf: disable
    AttrClass = lang.mkclass(name, (Attr, PrimType))
    AttrClass.PrimType = PrimType

    locals()[name] = AttrClass
    Attr.infer_types[PrimType] = AttrClass


def xget(obj, idx):
    "Does foo"
    if lang.isprimitive(obj):
        return obj
    if callable(obj):
        return obj(idx)  # ???

    def attempt_many():
        "Does attempt_many"
        vars_ = lang.pubvars(obj)
        attrs = fnmatch.filter(vars_, idx)
        if funcy.is_seqcoll(obj) or isinstance(obj, set):
            return attrs
        else:
            return [Attr.infer(attr, xget(obj, attr)) for attr in attrs]

    try:
        try:
            return obj[lang.maybe_int(idx)]
        except KeyError:
            return attempt_many()
    except TypeError:
        try:
            return getattr(obj, idx)
        except AttributeError:
            return attempt_many()


class AttrQuery(tuple):
    def __new__(cls, parts, *params, **opts):
        "Does __new__"
        pt, = lang.primbases(cls)
        instance = pt.__new__(cls, parts, *params, **opts)
        return instance

    @staticmethod
    def from_text(text):
        return AttrQuery(tuple(text.split('.')))


def idig(obj, path):
    "Recursive query for attributes from `obj` by a sequence spec."
    key = path.pop(0)
    point = xget(obj, key)
    if path:
        return idig(point, path)
    else:
        return point


def dig(obj, path):
    "Recursive query for attributes from `obj` by string spec."

    return idig(obj, path.split('.'))



def binstr(value, width=8):
    # type: (int, int) -> str
    prefixd = width + 3
    fmt = "#{}b".format(prefixd)
    return format(value, fmt).strip().replace('0b', '').zfill(width)


def snv(value, width=8):
    # type: (int, int) -> str
    return '  '.join(
        idx % 2 == 2 and ch + ' ' or ch for idx, ch in enumerate(binstr(value))
    )


def hexdump(bytes_):
    # type: (bytes) -> str
    return ' '.join(format(ord(x), 'x').zfill(2) for x in bytes_)


class ops:
    """Functions for bitwise operations. For use in e.g. generator
    expressions, `map()`, `filter()` etc.

    """

    @staticmethod
    def shiftl(steps, val):
        # type: (int, int) -> int
        "Shift left as a function. For use in functional patterns."
        return val << steps

    @staticmethod
    def shiftr(steps, val):
        # type: (int, int) -> int
        return val >> steps



# yapf

from typing import Any, Callable, Union, Collection

from collections.abc import Iterable

import funcy
from funcy import flow


class PipingExperiment(object):
    """An experiment pushing the pipe concept a bit longer. Not sure where
    I'm going with this but curious.

    """
    def __init__(self, seed):
        # type: (Any) -> PipingExperiment
        self.result = seed

    def __or__(self, step):
        # type: (Callable) -> PipingExperiment
        self.result = step(self.result)
        return self


P = PipingExperiment

IT = type(int)
itrt = Iterable

NarrowingPredicate = Union[Any, Callable[[Any], bool]]


class Narrowable(object):
    def narrow(self, pred):
        # type: (NarrowingPredicate) -> Narrowable

        if callable(pred):
            # Narrow by a callabale predicate
            pred_ = flow.ignore(Exception, False)(pred)
            return narrowable(self.base([el for el in self if pred_(el)]))

        else:
            # Narrow by exact Match
            return narrowable(self.base([el for el in self if pred == el]))

    def __getitem__(self, idx):
        # type: (Any) -> Any
        "Finds individual node in itself or searches.."
        try:
            return self.__class__(self.base.__getitem__(self, idx))
        except AttributeError:
            return self.narrow(idx)
        except TypeError:
            return self.narrow(idx)
        except KeyError:
            return self.narrow(idx)
        except IndexError:
            return self.narrow(idx)


def narrowable(src):
    # type: (Union[Any, Collection]) -> Narrowable
    "Creates the base class an initializes a Narrowable collection."

    class Narrower(Narrowable, src.__class__):
        base = src.__class__

    return Narrower(src)



# yapf

from functools import wraps, partial

from inspect import ismethod
from typing import Callable


class FSM(object):
    def __init__(self, initial):
        self.state = initial

    def methods(self):
        attr = partial(getattr, self)
        return (attr(n) for n in dir(self) if ismethod(attr(n)))

    def transition(self, state, *args, **kwargs):
        desc = f"{self.state}->{state}"

        def istransit(m: Callable) -> bool:
            return getattr(m, 'transits', '') == desc

        return (r for r in (transit for transit in (m(state, *args, **kwargs)
                                                    for m in self.methods()
                                                    if istransit(m))))


def state(name):
    def state_decorator(method):
        method.state = name

        @wraps(method)
        def event_handler(machine, *_, **__):
            if machine.state == method.state:
                return method(machine, *_, **__)
            else:
                # print(f"drop {method}, not it's state")
                return None

        return event_handler

    return state_decorator


def transits(desc):
    if '->' in desc:
        current, target = desc.split('->')
    elif '<-' in desc:
        current, target = desc.split('<-')

    def transit_decorator(method):
        def transit_handler(machine, newstate, *_, **__):
            result = method(machine, *_, **__)
            machine.state = target
            return result

        transit_handler.transits = desc
        return transit_handler

    return transit_decorator



# yapf

import sys
import io
from typing import Any, Tuple, List, Callable, Optional
import pdb
import traceback

import distutils.cmd

import funcy  # type: ignore
import pytest  # type: ignore

from altered import state  # type: ignore

import pycodestyle  # type: ignore
import mypy.api  # type: ignore

from flake8.api import legacy as flake8  # type: ignore


class fixture(object):
    @staticmethod
    def params(namelist: str, *values: Any) -> Any:
        "Does params"
        return pytest.mark.parametrize(namelist, values)


class ReviewProject(distutils.cmd.Command):  # pragma: nocov
    user_options: List[str] = []

    def initialize_options(self: 'ReviewProject') -> None:
        pass

    def finalize_options(self: 'ReviewProject') -> None:
        pass

    @staticmethod
    def lint() -> Tuple[int, str, str]:
        print('     flake8...')
        out, err = io.StringIO(), io.StringIO()
        guide = flake8.get_style_guide()
        with state(sys, stdout=out, stderr=err):
            report = guide.check_files('.')
        return (report.total_errors, out.getvalue(), err.getvalue())

    @staticmethod
    def style() -> Tuple[int, str, str]:
        """Runs pycodestyle, unpleasantly white-boxy call.  ..but it doesn't
         seem to be written for integration, so:

        """
        print('     pycodestyle...')
        code = 0
        warn, err = io.StringIO(), io.StringIO()
        with state(sys, argv=[], stdout=warn, stderr=err):
            try:
                pycodestyle._main()
            except SystemExit as exc:
                code = exc.code
        return code, warn.getvalue(), err.getvalue()

    @staticmethod
    def types() -> Tuple[int, str, str]:
        "Runs MyPy, the standard Python type checker."
        print('     MyPy...')
        warn, err, code = mypy.api.run(['.'])
        return code, warn, err

    @staticmethod
    def separator_line() -> None:
        "Prints a separating line of 79 characters wide."
        print()
        print(79 * '-')

    def run(self: 'ReviewProject') -> None:
        "Runs the setup task `'review'`."
        reports = {
            'flake8': ReviewProject.lint(),
            'pycodestyle': ReviewProject.style(),
            'mypy': ReviewProject.types(),
        }

        for tool in reports:
            code, warn, err = reports[tool]
            if code != 0:
                ReviewProject.separator_line()
                print(f'{tool}: code {code}')
                print('WARN')
                print(warn)
                print('ERROR')
                print(err)

        issues = sum(funcy.walk_values(lambda x: x[0], reports).values())
        if issues > 0:
            ReviewProject.separator_line()
            print()
            print(f'{issues} issues found.')


def hook_uncatched(pm_func: Optional[Callable] = None
                   ) -> None:  # pragma: nocov
    """Installs an exception hook that triggers a debugger post-mortem on
    unhandled exceptions.

    """

    if pm_func is None:
        pm_func = pdb.post_mortem

    def unhandled(type_: Any, value: Any, tb: Any) -> None:
        """Catches unhandled exceptions, print stack trace and executes
        debugger post-mortem entry point.

        """
        traceback.print_exception(type_, value, tb)
        print()  # lf
        pm_func(tb)

    sys.excepthook = unhandled
    return unhandled



# yapf

import funcy
import types
import numbers
import copy
from functools import singledispatch
from pysistence import Expando
from typing import (Any, Mapping, List, Tuple, Iterable, Generator, Callable,
                    Union)
import inspect
from . import pipelib
import itertools

from functools import wraps, update_wrapper

import operator as ops

PRIMTYPES = {int, bool, float, str, set, list, tuple, dict}
LISTLIKE = {set, list, tuple}

textual = funcy.isa(str)
numeric = funcy.isa(numbers.Number)
isint = funcy.isa(int)
isdict = funcy.isa(dict)
isgen = funcy.isa(types.GeneratorType)


def unfold_gen(x: Generator[Any, None, None],
               cast: type = tuple) -> Iterable[Any]:
    """Quick recursive unroll of possibly nested (uses funcy library under
    the hood)

    """
    return cast(funcy.flatten(x, isgen))


def typename(x: Any) -> str:
    """Safely get a type name from parameter `x`. If `x == None` returns
    `'None'`

    """
    return x is None and 'None' or x.__class__.__name__


class XE(Expando):
    def __getitem__(self: 'XE', attr: str) -> Any:
        "Support for getting attributes as named string indexes"
        return getattr(self, attr)

    def iteritems(self: 'XE') -> Mapping[str, Any]:
        "Does iteritems"
        return self.to_dict().items()

    def get(self, name, default=None):
        return self.to_dict().get(name, default)


def pubvars(obj: Any) -> Iterable:
    "Returns all public variables except methods"
    if isdict(obj):
        return obj.keys()
    if funcy.is_seqcoll(obj) or isinstance(obj, set):
        return copy.copy(obj)
    else:
        return [
            attr for attr in dir(obj)
            if not attr.startswith('__') and not callable(getattr(obj, attr))
        ]


def isprimitive(obj):
    "Determines if a value belongs to a primitive type (= {numbers, strings})"

    if numeric(obj):
        return True
    elif textual(obj):
        return True

    return False


def isprim_type(type_):
    "Does primtype"
    return True if type_ in PRIMTYPES else False


def tolerant_or_original(Exc, fn):
    """Returns a function that will allow the Exception(s) in `Exc` to
    occur."""
    def invoke(obj) -> Any:
        "Does invoke"
        try:
            return fn(obj)
        except Exc:
            return obj

    return invoke


def coerce_or_same(T: Any) -> str:
    "Special case of `tolerant_or_original()` for type coercion."
    "Does coerce_or_same"
    return tolerant_or_original((TypeError, ValueError, AttributeError), T)


maybe_int = coerce_or_same(int)


def methdispatch(func):
    """Thanks Zero Piraeus!

    https://stackoverflow.com/a/24602374/288672
    """
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)

    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


class Undefined:
    "Marker class for undefined values."


def primbases(cls):
    "Does primbase"
    return [T for T in cls.__bases__ if isprim_type(T)]


def bind_methods(Base, instance):
    "Does bind_methods"
    for name, fn in instance.__bind__:
        setattr(Base, name, types.MethodType(fn, instance))
    return instance


class Base(object):
    __bind__: List[Tuple[Callable, Any]] = []

    def __init__(self, *params, **opts):
        bind_methods(self.__class__, self)

        def bind_method_on_self(self: Any, fn: Callable):
            "Does bound_on_instance"

            name = fn.__name__

            @wraps(fn)
            def callit(*params: Any, **opts: Any) -> None:
                "Does callit"
                return fn(self, *params, **opts)

            setattr(self, name, callit)
            return callit

        self.method = types.MethodType(bind_method_on_self, self)

    @staticmethod
    def __wrapper(fn: Callable, Cls: Any = None, name: str = None):
        "Does __wrapper"
        if name is None:
            name = fn.__name__

        @wraps(fn)
        def do_call(*params, **opts):
            if Cls is None:
                return fn(*params, **opts)
            else:
                return fn(*((Cls, ) + params), **opts)

        if Cls is None:
            setattr(Base, name, do_call)
        else:
            setattr(Cls, name, do_call)

        return do_call

    __classmethod__ = classmethod

    @classmethod
    def classmethod(cls, fn):
        return Base.__wrapper(fn, Cls=cls)

    @staticmethod
    def staticmethod(fn: Callable) -> Callable:
        return Base.__wrapper(fn)

    @__classmethod__
    def method(Subclass: Any, fn: Callable) -> Callable:
        "Creates a method from the decorated function `fn`"
        setattr(Subclass, fn.__name__, types.MethodType(fn, Subclass))
        return fn


def mkclass(name: str, bases: Tuple = (), **clsattrs: Any) -> Any:
    "Does mkclass"

    Gen = type(name, (Base, ) + bases, clsattrs)
    return Gen


def arity(fn: Callable) -> int:
    "Returns the number of arguments required by `fn`."
    return len(inspect.signature(fn).parameters)


always_tup = funcy.iffy(funcy.complement(funcy.is_seqcont), lambda x: (x, ))


class Piping(pipelib.BasePiping):
    """Piping objects is for (ab)using Python operator overloading to
    build small pipeline-DSL's.

    The most basic one will simply refuse to do anything - you have to
    give it instructions/permissions on everything it's made for ;-).

    """
    class Fresh(object):
        "Marker for Piping instances that never has been run"
        pass

    class Executed(object):
        "Marker for Piping instances that has been run"
        pass

    def __init__(self,
                 seed: Union[tuple, Any] = (),
                 kind: Callable = map,
                 format: Callable[[Any, None, None], Any] = funcy.identity):
        "Initialize a new Piping object."
        self.reset()
        self.format = format
        self.kind = kind

        self.seed = always_tup(seed)

    def reset(self):
        "Restart this Piping object as if it were new."
        self.cursor = ()
        self.ops = ()
        self.results = {}
        self.last_result = ()

    def sum(self) -> None:
        "Does show"
        print(self.__class__.__name__)
        textual = (
            f'kind = {self.kind}, state = {self.state}, seed = {self.seed},'
            f'cursor = {self.cursor}, last result = {self.last_result}')
        print(textual)

    @property
    def now(self):
        return self.last_result

    def fncompose(self, stepf: Callable[[Any, None, None], Any],
                  x: Any = None) -> 'Piping':
        self.queue(stepf, x)
        return self

    def queue(self, stepf: Callable[[Any, None, None], Any],
              *x: Any) -> 'Piping':
        "Does queue"
        self.ops = (*self.ops, ((stepf, x)))
        return self

    def run(self, seed, *x: Any) -> Any:
        "Does do"
        self.cursor = seed
        for op, operands in self.ops:
            if operands:
                self.cursor = op(*(self.cursor, *operands))
            else:
                self.cursor = op(self.cursor)
        return self.cursor

    @property
    def state(self):
        if self.cursor == () and self.last_result == ():
            return Piping.Fresh
        elif self.last_result != ():
            return Piping.Executed

        raise ValueError('Piping: undeterminable state')

    def calc_fmt_param(self, kind, res, params, case):
        # case = (self.kind, ) + self.seed + self.ops
        if kind is filter and self.state == Piping.Fresh:
            # This case is logically obsolete but good for readability.
            return (res, )
        elif self.kind is filter:
            # Always compares against call parameters
            return params
        elif kind is map or kind == 'pipe':
            return (res, )

        raise TypeError('Piping: unknown out parameter scenario')

    def __call__(self, *params: Any) -> Any:
        """Treating the Pipe as a function calculates the Pipe's result and
        returns it passed through the return formatting function.

        Unary calls for pipelined function composition behavoiour,
        binary calls to combine a value with the result of the pipe
        operations (map/filter/reduce).

        """

        # Calculate case
        case = (self.kind, ) + self.seed + self.ops + params

        # Decide on operands for this run
        if self.state == Piping.Fresh:
            # When the pipeline must be run, prefer parameter but
            # otherwise seed
            operands = params or self.seed
        else:
            # When a result exist, prefer seed
            operands = self.seed or params

        # An incorrectly handled pipe is a possibility here
        if not operands:
            raise ValueError('Piping: fatal/undeterminable operands')

        try:
            # Stored, grab this case from results storage
            res = self.results[case]
        except KeyError:
            # This case has not been stored yet
            res = self.run(*operands)

            # On success:
            # Store last result in tuple form
            self.last_result = always_tup(res)
            # Also store the Piping case.
            self.results[case] = res

        # Decide on return format
        format_params = self.calc_fmt_param(self.kind, res, params, case)
        # Invoke return format function with parameter(s) from above
        return self.format(*format_params)


class ComposePiping(Piping):
    """Common usage of Piping - the >> operator is the simplest possible
    function composition. I decided against the | operator since it's
    needed for logical pipes.

    Most implementations will probably be based of this.

    """
    def __rshift__(self, stepf: Callable[[Any, None, None], Any]) -> None:
        "Bitwise OR as simple function composition"
        return self.queue(stepf)


class CountPiping(Piping):
    def __add__(self, value) -> Piping:
        "Add operation"
        return self.queue(ops.add, value)

    def __sub__(self, value) -> Piping:
        "Add operation"
        return self.queue(ops.sub, value)

    def __mul__(self, value) -> Piping:
        "Add operation"
        return self.queue(ops.mul, value)

    def __div__(self, value) -> Piping:
        "Add operation"
        return self.queue(ops.truediv, value)


def PNot(value_or_stepf: Union[Callable, Any]) -> bool:
    "Unary negate for `LogicPiping`."
    if callable(value_or_stepf):
        stepf = value_or_stepf
    else:
        stepf = funcy.identity
    return funcy.complement(stepf)


rcurry = funcy.rcurry


class LogicPiping(Piping):
    def __init__(self,
                 seed: Union[tuple, Any] = (),
                 kind: Callable = map,
                 format: Callable[[Any, None, None], Any] = funcy.identity):
        "Sensible default, override for advanced use."
        super().__init__(seed, kind, format)
        self.truthy = []
        self.conjunctions = 0

    def logically(self, stepf, conjunction):
        self.counter = itertools.count(1)
        if conjunction:
            self.conjunctions += 1

        def gate(x):
            if stepf(x):
                self.truthy.append(x)
            else:
                self.truthy.append(False)
            return x

        return self.queue(gate)

    def __call__(self, *params: Any) -> Any:
        self.truthy = []
        super().__call__(*params)
        passed = funcy.compact(self.truthy)
        if len(passed) == self.conjunctions:
            return self.format(passed[-1])
        else:
            return False

    def __eq__(self, value) -> None:
        "Does __ge__"
        return self.logically(rcurry(ops.eq)(value), True)

    def __neg__(self, stepf) -> None:
        "Does __ge__"
        return self.logically(funcy.complement(stepf), False)

    def __ne__(self, value) -> None:
        "Does __ge__"
        return self.logically(rcurry(ops.ne)(value), True)

    def __ge__(self, value) -> None:
        "Does __ge__"
        return self.logically(rcurry(ops.ge)(value), True)

    def __gt__(self, value) -> None:
        "Does __ge__"
        return self.logically(rcurry(ops.gt)(value), True)

    def __and__(self, stepf: Callable[[Any, None, None], Any]) -> Piping:
        return self.logically(stepf, True)

    def __floordiv__(self, stepf: Callable[[Any, None, None], Any]) -> Piping:
        "Aestetics only."
        return self.__and__(stepf)

    def __or__(self, stepf: Callable[[Any, None, None], Any]) -> Piping:
        return self.logically(stepf, False)



# yapf

import pprint
import textwrap


def nice(ob, compress=2):
    # type: (ob) -> None
    "Does nice"
    return textwrap.dedent(pprint.pformat(ob)).replace(compress * ' ', ' ')


def binstr(value, width=8):
    prefixd = width + 3
    fmt = "#{}b".format(prefixd)
    return format(value, fmt).strip().replace('0b', '').zfill(width)


def nv(value, width=8):
    return binstr(value, width)


def snv(value, width=8):
    return '\n'.join(('  '.join(idx % 2 == 2 and ch + ' ' or ch
                                for idx, ch in enumerate(nv(value, width))),
                      '0  1  2  3  4  5  6  7'))


def visual_bits_print(value, width=8):
    out = snv(value, width)
    print(out)
    return out



 #!/usr/bin/env python
# yapf

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os
import kingston
import tempfile
import subprocess

install = (
    "funcy>=1.10.2",
    "pysistence>=0.4.1",
    "patterns>=0.3",
    "jsonpickle>=1.2",
)  # yapf: disable

develop = (
    "pytest>=5.0.1",
    "hypothesis>=4.24.3",
    "altered_states>=1.0.9",
    "pycodestyle>=2.5.0",
    "mypy>=0.730",
    "pyflakes>=2.1.1",
    "flake8",
)  # yapf: disable

try:
    # make setup.py review available only to fully dev-capable
    # interpreter instances
    import mypy
    assert mypy

    from kingston.testing import ReviewProject
    cmdclass = {'review': ReviewProject}
except ImportError:
    cmdclass = {}

# Try to convert README to RST using pandoc (allowed to fail).
long_desc = ''
try:
    tmpdir = tempfile.mkdtemp()
    outpath = os.path.join(tmpdir, 'kingston-README.rst')
    subprocess.check_call(['pandoc', 'README.org', '-o', outpath])
    with open(outpath) as fp:
        long_desc = fp.read()
except Exception:
    long_desc = ''  # nothing to do.

setup(
    name='kingston',
    cmdclass=cmdclass,
    version=kingston.__version__,
    description="Some Python nicieties",
    long_description=long_desc,
    packages=('kingston', ),
    author='Jacob Oscarson',
    author_email='jacob@414soft.com',
    install_requires=install,
    extras_require={
        'test': install + develop,
    },
    url='https://www.414soft.com/kingston',
    license='MIT',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



# yapf


class BasePiping(object):  # type: ignore
    ...


class BasePiping(object):
    def __add__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __sub__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement -')

    def __mul__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement *')

    def __matmul__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement @')

    def __truediv__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement /')

    def __floordiv__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement //')

    def __mod__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement %')

    def __divmod__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement divmod')

    def __pow__(self, other, modulo=None):
        raise NotImplementedError('Does not implement pow')

    def __lshift__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement <<')

    def __rshift__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement >>')

    def __and__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement &')

    def __xor__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement ^')

    def __or__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement |')

    def __radd__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand +')

    def __rsub__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand -')

    def __rmul__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand *')

    def __rmatmul__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand @')

    def __rtruediv__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand /')

    def __rfloordiv__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement rightand %')

    def __rmod__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand %')

    def __rdivmod__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand divmod')

    def __rpow__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand pow')

    def __rlshift__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand +')

    def __rrshift__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand +')

    def __rand__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand +')

    def __rxor__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand +')

    def __ror__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement righthand +')

    def __iadd__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +=')

    def __isub__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement -=')

    def __imul__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement *=')

    def __imatmul__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __itruediv__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __ifloordiv__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __imod__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement %=')

    def __ipow__(self, other, modulo=None):
        raise NotImplementedError('Does not implement **=')

    def __ilshift__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement <<=')

    def __irshift__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement >>=')

    def __iand__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __ixor__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __ior__(self, other) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __neg__(self) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __pos__(self) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __abs__(self) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __invert__(self) -> BasePiping:
        raise NotImplementedError('Does not implement +')

    def __int__(self) -> BasePiping:
        return int(self.result)

    def __float__(self) -> BasePiping:
        raise float(self.result)



# yapf

from typing import Any, Iterable, Callable, Union
import itertools
import funcy  # type: ignore
from funcy import flow
from functools import partial
from . import lang

import inspect


class Mismatch(ValueError):
    pass


class Conflict(TypeError):
    pass


class Match(dict):
    """Multiple dispatch as a callable subclass of `dict`:

    Define a mapping of types. Call for an instance, the parameter
    type should be mapped by `type(obj)` returning a callable that
    will further process the instance.

    """
    def checkpred(self, key: Any, value: Any) -> None:
        if key in self:
            confl = self[key]
            raise Conflict(f'Pattern {key} had a previous conflict '
                           f'{key}={confl}')

    @staticmethod
    def wccmp(needle: Any, match: Any):
        if match is Any:
            return True
        elif callable(match):
            return flow.silent(match)(needle)
        else:
            return needle == match

    def case(*params: Any) -> Callable:
        """Decorator to add a function. The types of the parameters. The types
        that will be matched is taken from the signature of the
        decorated function.

        """
        self, fn = params
        disp = tuple(arg.annotation
                     for arg in inspect.signature(fn).parameters.values())

        self[disp] = fn

        return fn

    def type_match(self, call: Callable, *params: Any, **opts: Any) -> Any:
        "Does type_match"
        fparams = tuple(funcy.flatten(params))
        T = tuple(type(p) for p in fparams)
        if len(T) == 1:
            call = self[T[0]]
        else:
            scan = partial(self.wc_scan, call, fparams, **opts)
            call = funcy.fallback(lambda: self[T],
                                  lambda: self[type(T)],
                                  scan)  # yapf: disable
        return call() if lang.arity(call) == 0 else call(*fparams, **opts)

    def wc_scan(self, call: Callable, values: Iterable, **opts: Any):
        for preds, fn in self.items():
            if callable(preds) and flow.silent(preds)(*values, **opts):
                call = fn
            elif funcy.iterable(preds):
                for pred in preds:
                    if all(
                            VMatch.wccmp(a, b)
                            for a, b in itertools.zip_longest(values, preds)):
                        call = fn
        return call

    def __call__(self, *params: Any,
                 **opts: Any) -> Union[Iterable[Any], Iterable]:
        "Return the value keyed by the type of parameter `obj`"
        call = flow.raiser(Mismatch,
                           f"Match: no match for parameters {params!r}")
        return self.type_match(call, *params, **opts)

    def __setitem__(self, key: Any, value: Any) -> None:
        "Insert new matching, but check for pre-existing matches first."
        self.checkpred(key, value)
        super().__setitem__(key, value)


class VMatch(Match):
    def checkpred(self, key: Any, value: Any) -> None:
        super().checkpred(key, value)
        try:
            self(key)
            raise Conflict(f'Pattern {key} could already be type matched.')
        except Mismatch:
            pass

    def value_match(self, call: Callable, *params, **opts):
        """
        Match parameters by value
        """
        fparams = tuple(funcy.flatten(params))
        try:
            # Exact match
            call = self[fparams]
        except KeyError:
            for preds, fn in self.items():
                if callable(preds) and flow.silent(preds)(*fparams, **opts):
                    call = fn
                elif funcy.iterable(preds):
                    call = self.wc_scan(call, fparams, **opts)

        return call() if lang.arity(call) == 0 else call(*fparams, **opts)

    def __call__(self, *params: Any,
                 **opts: Any) -> Union[Iterable[Any], Iterable]:
        """
        Handle call to this instance.
        """
        call = flow.raiser(Mismatch,
                           f"Value Match: no match for parameters {params!r}")
        return self.value_match(call, *params, **opts)
